"""
Basic universal log configuration for project

in each file you want to log,
import set_log and configure with
`log = set_log(__name__)`
"""
import logging.config
from logging import Logger


def set_log(name: str) -> Logger:
    """
    :param name: usually the module path __name__
    :return: Configured Logger
    """
    log = logging.getLogger(name)
    try:
        logging.config.fileConfig(fname="logging.conf", disable_existing_loggers=True)
    except KeyError:
        pass
    return log



import unittest

from duneapi.api import DuneAPI
from duneapi.dashboard import DuneDashboard


class TestDashboard(unittest.TestCase):
    def test_update(self):
        dashboard = DuneDashboard.from_file(
            api=DuneAPI.new_from_environment(),
            filename="./example/dashboard/_config.json",
        )
        dashboard.update()
        query2 = dashboard.queries[1]
        self.assertEqual(dashboard.api.fetch(query2), [{"val": "1337"}])

        # Modify the query (by changing the parameter) and update dashboard
        query2.parameters[0].value = 10
        dashboard.update()
        self.assertEqual(dashboard.api.fetch(query2), [{"val": "10"}])


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from duneapi.api import DuneAPI
from duneapi.dashboard import DuneDashboard
from duneapi.types import DashboardTile, DuneQuery


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.user = "TestUser"
        self.dune = DuneAPI(username=self.user, password="")
        self.query_path = "./example/dashboard"
        self.queries = [
            {
                "id": 1,
                "name": "Example 1",
                "query_file": "query1.sql",
                "network": "mainnet",
                "requires": "./example/dashboard/base_query.sql",
            },
            {
                "id": 2,
                "name": "Example 2",
                "query_file": "query2-polygon.sql",
                "network": "polygon",
            },
        ]
        self.valid_input = json.loads(
            json.dumps(
                {
                    "meta": {
                        "name": "Demo Dashboard",
                        "user": self.user,
                        "query_path": self.query_path,
                    },
                    "queries": self.queries,
                }
            )
        )

    def test_constructor(self):
        dashboard = DuneDashboard.from_json(self.dune, self.valid_input)
        expected_tiles = [
            DashboardTile.from_dict(q, self.query_path) for q in self.queries
        ]
        expected_queries = [DuneQuery.from_tile(t) for t in expected_tiles]

        self.assertEqual(dashboard.name, "Demo Dashboard")
        self.assertEqual(dashboard.url, f"https://dune.xyz/{self.user}/Demo-Dashboard")
        self.assertEqual(dashboard.queries, expected_queries)

    def test_user_assertion(self):
        minimal_input = {
            "meta": {
                "name": "Demo Dashboard",
                "user": "WrongUser",
            },
            "queries": [],
        }
        # TODO - figure out how to use self.assertLogs
        with self.assertRaises(ValueError) as err:
            DuneDashboard.from_json(self.dune, minimal_input)
        self.assertEqual(
            str(err.exception),
            f"Attempt to load dashboard queries for invalid user WrongUser != {self.user}.",
        )

    def test_duplicate_query(self):
        query_path = "./tests/queries"
        query_file = "test_query.sql"
        minimal_input = {
            "meta": {
                "name": "Demo Dashboard",
                "user": self.user,
                "query_path": query_path,
            },
            "queries": [
                {
                    "id": 1,
                    "name": "Example 1",
                    "query_file": query_file,
                    "network": "gchain",
                },
                {
                    "id": 2,
                    "name": "Example 2",
                    "query_file": query_file,
                    "network": "gchain",
                },
            ],
        }

        with self.assertLogs("updated_networks_dune.dashboard", level="WARN") as cm:
            DuneDashboard.from_json(self.dune, minimal_input)
        self.assertEqual(
            cm.output,
            [
                "WARNING:updated_networks_dune.dashboard:Duplicate Query Detected [(\"select 10 - '{{IntParameter}}' as value\", <Network.GCHAIN: 6>)]"
            ],
        )


if __name__ == "__main__":
    unittest.main()



"""Dashboard management class"""
from __future__ import annotations

import argparse
import json
import os
import time

from typing import Any

from .api import DuneAPI
from .constants import FIND_DASHBOARD_POST, FIND_QUERY_POST
from .logger import set_log
from .types import DuneQuery, DashboardTile, Post, Network, QueryParameter
from .util import duplicates

BASE_URL = "https://dune.xyz"
log = set_log(__name__)


class DuplicateQueryError(Exception):
    """Basic extension of Exception class"""


class DuneDashboard:
    """
    A Dune Dashboard consists of a family of queries
    Primary functionality is to update all of them
    without having to manually click refresh.
    """

    name: str
    url: str
    queries: list[DuneQuery]
    api: DuneAPI

    def __init__(
        self, api: DuneAPI, name: str, slug: str, user: str, queries: list[DuneQuery]
    ):
        dupes = duplicates([(q.raw_sql, q.network) for q in queries])
        if dupes:
            log.warning(f"Duplicate Query Detected {dupes}")
            # raise DuplicateQueryError(dupes)

        if api.username != user:
            raise ValueError(
                f"Attempt to load dashboard queries for invalid user {user} != {api.username}."
            )
        self.name = name
        self.slug = slug
        self.url = "/".join([BASE_URL, user, slug])
        self.queries = list(queries)
        self.api = api

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DuneDashboard):
            return NotImplemented
        equality_conditions = [
            self.name == other.name,
            self.slug == other.slug,
            self.url == other.url,
            set(self.queries) == set(other.queries),
        ]
        log.debug(f"Equality conditions: {equality_conditions}")
        return all(equality_conditions)

    @classmethod
    def from_file(cls, api: DuneAPI, filename: str) -> DuneDashboard:
        """Constructs Dashboard from configuration file"""
        with open(filename, "r", encoding="utf-8") as config:
            return cls.from_json(api=api, json_obj=json.loads(config.read()))

    @classmethod
    def from_dune(
        cls, api: DuneAPI, dashboard_slug: str, save_config: bool = True
    ) -> DuneDashboard:
        """
        Initialized instance by fetching existing Dashboard from Dune.
        When save_config is True, Saves dashboard config files in ./out
        """
        post_data = {
            "operationName": "FindDashboard",
            "variables": {
                "session_id": {"_eq": 0},
                "user": api.username,
                "slug": dashboard_slug,
            },
            "query": FIND_DASHBOARD_POST,
        }
        response = api.post_dune_request(
            Post(data=post_data, key_map={"dashboards": {"visualization_widgets"}})
        )
        meta = response.json()["data"]["dashboards"][0]
        widgets = meta["visualization_widgets"]
        queries = set()
        for widget in widgets:
            query_data = widget["visualization"]
            query_id = query_data["query_details"]["query_id"]

            post = Post(
                data={
                    "operationName": "FindQuery",
                    "variables": {"session_id": {"_eq": 0}, "id": query_id},
                    "query": FIND_QUERY_POST,
                },
                key_map={},
            )
            response = api.post_dune_request(post)
            log.debug(f"Received Response {response.json()}")
            query_data = response.json()["data"]["queries"][0]
            # Filtering out queries that are not owned by logged-in user.
            if query_data["user"]["name"] == api.username:
                queries.add(
                    DuneQuery(
                        name=query_data["name"],
                        description=query_data["description"],
                        raw_sql=query_data["query"],
                        network=Network(query_data["dataset_id"]),
                        parameters=[
                            QueryParameter.from_dict(p)
                            for p in query_data["parameters"]
                        ],
                        query_id=query_data["id"],
                    )
                )
            else:
                log.info(
                    f'Ignoring dashboard query from user {query_data["user"]["name"]}'
                )

        dashboard_owner = meta["user"]["name"]
        assert dashboard_owner == api.username, "Dashboard not owned by user"

        if save_config:
            cls.dump_config(
                name=meta["name"],
                owner=dashboard_owner,
                slug=dashboard_slug,
                queries=list(queries),
            )
        return cls(
            api=api,
            name=meta["name"],
            slug=dashboard_slug,
            queries=list(queries),
            user=dashboard_owner,
        )

    @staticmethod
    def dump_config(name: str, owner: str, slug: str, queries: list[DuneQuery]) -> None:
        """
        Writes Dashboard Configuration to files.
        Specifically to ./out/Dashboard-Slug
        """
        out_dir = f"./out/{slug}"
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        queries_seen = {}
        with open(f"{out_dir}/_config.json", "w", encoding="utf-8") as config_file:
            query_dicts = []
            for query in queries:
                query_file = f"{query.name.lower().replace(' ', '-')}.sql"
                query_config = {
                    "id": query.query_id,
                    "name": query.name,
                    "description": query.description,
                    "query_file": query_file,
                    "network": str(query.network),
                    "parameters": [p.to_dict() for p in query.parameters],
                }
                query_dicts.append(query_config)

                if query.raw_sql not in queries_seen:
                    # Whenever we encounter a new SQL query, write to file
                    query_path = f"{out_dir}/{query_file}"
                    with open(query_path, "w", encoding="utf-8") as q_file:
                        q_file.write(query.raw_sql.strip("\n") + "\n")
                    queries_seen[query.raw_sql] = query_file
                else:
                    # If already seen use the already existing file.
                    query_config["query_file"] = queries_seen[query.raw_sql]

            config_dict = {
                "meta": {
                    # Dashboards can be renamed but the slug doesn't change.
                    "name": name,
                    "slug": slug,
                    "user": owner,
                    "query_path": out_dir,
                },
                "queries": query_dicts,
            }
            config_file.write(
                json.dumps(config_dict, indent=2, default=str).strip("\n") + "\n"
            )

    @classmethod
    def from_json(cls, api: DuneAPI, json_obj: dict[str, Any]) -> DuneDashboard:
        """Constructs Dashboard from json file"""
        meta, queries = json_obj["meta"], json_obj["queries"]
        # TODO - tiles could be phased out of this program.
        tiles = [DashboardTile.from_dict(q, meta["query_path"]) for q in queries]
        queries = [DuneQuery.from_tile(tile) for tile in tiles]
        name = meta["name"]
        return cls(
            api=api,
            name=name,
            # Dashboards can be renamed, but slug is permanent.
            # So we chose slug with priority and use name otherwise.
            slug=meta.get("slug", name.replace(" ", "-")),
            user=meta["user"],
            queries=queries,
        )

    def update(self) -> None:
        """Creates a dune connection and updates/refreshes all dashboard queries"""
        for tile in self.queries:
            self.api.initiate_query(tile)
            retries, success = 0, False
            # TODO - this is not a very good way of retrying
            # Should create custom errors (instead of using Runtime everywhere)
            while not success and retries < 3:
                try:
                    self.api.execute(tile.query_id, tile.parameters)
                    success = True
                except RuntimeError as err:
                    retries += 1
                    sleep_time = 2 * retries
                    log.warning(
                        f"Query execution failed due to {err}. "
                        f"Sleeping {sleep_time} seconds and trying again."
                    )
                    time.sleep(sleep_time)
            if not success:
                raise SystemExit("Failed to update dashboard, max retries exceeded!")

    def __str__(self) -> str:
        names = "\n".join(
            f"  {q.name}: {BASE_URL}/queries/{q.query_id}" for q in self.queries
        )
        return f'Dashboard "{self.name}": {self.url}\nQueries:\n{names}'


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Fetch Allocation Receipts for a list of accounts"
    )
    parser.add_argument(
        "--dashboard-slug",
        type=str,
        help="The hyphenated last part of the dashboard URL",
    )
    args = parser.parse_args()

    dune = DuneAPI.new_from_environment()
    dashboard = DuneDashboard.from_dune(dune, args.dashboard_slug)
    print("Updated", dashboard)



"""Sample Fetch script from DuneAnalytics"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from duneapi.file_io import write_to_csv, File
from duneapi.api import DuneAPI
from duneapi.types import Network, QueryParameter, DuneQuery, Address
from duneapi.util import open_query


@dataclass
class Record:
    """Arbitrary record with a few different data types"""

    string: str
    address: Address
    integer: int
    decimal: float
    time: datetime

    @classmethod
    def from_dict(cls, obj: dict[str, str]) -> Record:
        """Constructs Record from Dune Data as string dict"""
        return cls(
            string=obj["block_hash"],
            address=Address(obj["miner"]),
            integer=int(obj["number"]),
            decimal=float(obj["tx_fees"]),
            # Dune timestamps are UTC!
            time=datetime.strptime(obj["time"], "%Y-%m-%dT%H:%M:%S+00:00"),
        )


def fetch_records(dune: DuneAPI) -> list[Record]:
    """Initiates and executes Dune query, returning results as Python Objects"""
    sample_query = DuneQuery.from_environment(
        raw_sql=open_query("./example/query.sql"),
        name="Sample Query",
        network=Network.MAINNET,
        parameters=[
            QueryParameter.number_type("IntParam", 10),
            QueryParameter.date_type("DateParam", datetime(2022, 3, 10, 12, 30, 30)),
            QueryParameter.text_type("TextParam", "aba"),
        ],
    )
    results = dune.fetch(sample_query)
    return [Record.from_dict(row) for row in results]


if __name__ == "__main__":
    records = fetch_records(DuneAPI.new_from_environment())
    print("First result:", records[0])
    write_to_csv(data_list=records, outfile=File("example-output.csv"))



"""Utility methods to support Dune API"""
import collections
from datetime import datetime
from typing import Any, Hashable

DUNE_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def postgres_date(date_str: str) -> datetime:
    """Parse a postgres compatible date string into datetime object"""
    return datetime.strptime(date_str, DUNE_DATE_FORMAT)


def datetime_parser(dct: dict[str, Any]) -> dict[str, Any]:
    """
    Used as object hook in json loads method to parse postgres dates strings
    """
    for key, val in dct.items():
        if isinstance(val, str):
            try:
                dct[key] = datetime.strptime(val, DUNE_DATE_FORMAT)
            except ValueError:
                pass
    return dct


def open_query(filepath: str) -> str:
    """Opens `filename` and returns as string"""
    with open(filepath, "r", encoding="utf-8") as query_file:
        return query_file.read()


def duplicates(arr: list[Hashable]) -> list[Hashable]:
    """Detects and returns duplicates in array"""
    return [item for item, count in collections.Counter(arr).items() if count > 1]



import unittest
from unittest.mock import MagicMock

from requests import Response

from duneapi.response import (
    pre_validate_response,
    validate_and_parse_dict_response,
    validate_and_parse_list_response,
)


class TestOperations(unittest.TestCase):
    def setUp(self) -> None:
        self.response = Response()
        self.key_map = {"x": {"y"}}
        self.valid_dict_data = {"data": {"x": {"y": "z"}}}
        self.valid_list_data = {"data": {"x": [{"y": "z"}]}}

    def test_pre_validation_success(self):
        self.response.status_code = 200
        self.response.json = MagicMock(return_value=self.valid_dict_data)
        response_data = pre_validate_response(self.response, key_map=self.key_map)
        self.assertEqual(response_data, self.valid_dict_data["data"])

    def test_pre_validation_errors(self):
        self.response.status_code = 1
        with self.assertRaises(RuntimeError) as err:
            pre_validate_response(self.response, {})
        self.assertEqual(
            str(err.exception), "('Dune post failed with', <Response [1]>)"
        )

        self.response.status_code = 200
        self.response.json = MagicMock(return_value={"x": {"y": "z"}})
        with self.assertRaises(ValueError) as err:
            pre_validate_response(self.response, key_map={"x": {"y"}})
        self.assertEqual(
            str(err.exception), "response json {'x': {'y': 'z'}} missing 'data' key"
        )

        self.response.json = MagicMock(return_value={"data": {"query_errors": 5}})
        with self.assertRaises(RuntimeError) as err:
            pre_validate_response(self.response, {"x": {"y"}})
        self.assertEqual(str(err.exception), "Dune API Request failed with errors 5")

        self.response.json = MagicMock(return_value=self.valid_dict_data)
        with self.assertRaises(AssertionError) as err:
            pre_validate_response(self.response, key_map={"wrong key": {"y"}})
        self.assertEqual(
            str(err.exception),
            "got=dict_keys(['x']), expected=dict_keys(['wrong key'])",
        )

    def test_dict_validation_success(self):
        self.response.status_code = 200
        self.response.json = MagicMock(return_value=self.valid_dict_data)
        response_data = validate_and_parse_dict_response(
            self.response, key_map=self.key_map
        )
        self.assertEqual(response_data, self.valid_dict_data["data"])

    def test_list_validation_success(self):
        self.response.status_code = 200
        self.response.json = MagicMock(return_value=self.valid_list_data)
        response_data = validate_and_parse_list_response(
            self.response, key_map=self.key_map
        )
        self.assertEqual(response_data, self.valid_list_data["data"])

    def test_dict_validation_error(self):
        self.response.status_code = 200
        # Note that valid list data is not valid dict data
        self.response.json = MagicMock(return_value=self.valid_list_data)
        with self.assertRaises(AssertionError) as err:
            validate_and_parse_dict_response(self.response, key_map=self.key_map)
        self.assertEqual(str(err.exception), "Invalid response type <class 'list'>")

        partially_valid_dict_data = {"data": {"x": {"a": "b"}}}
        self.response.json = MagicMock(return_value=partially_valid_dict_data)
        with self.assertRaises(AssertionError) as err:
            validate_and_parse_dict_response(self.response, key_map=self.key_map)
        self.assertEqual(str(err.exception), "Fail dict_keys(['a']) != {'y'}")

    def test_list_validation_error(self):
        self.response.status_code = 200
        # Note that valid dict data is not valid list data
        self.response.json = MagicMock(return_value=self.valid_dict_data)
        with self.assertRaises(AssertionError) as err:
            validate_and_parse_list_response(self.response, key_map=self.key_map)
        self.assertEqual(str(err.exception), "Invalid response type <class 'dict'>")

        partially_valid_list_data = {"data": {"x": [{"a": "b"}]}}
        self.response.json = MagicMock(return_value=partially_valid_list_data)
        with self.assertRaises(AssertionError) as err:
            validate_and_parse_list_response(self.response, key_map=self.key_map)
        self.assertEqual(str(err.exception), "Fail dict_keys(['a']) != {'y'}")


if __name__ == "__main__":
    unittest.main()



"""
Post data associated with Dune Analytics API
"""
# pylint: disable-msg=R0801
FIND_DASHBOARD_POST = """
    query FindDashboard($session_id: Int_comparison_exp, $user: String!, $slug: String!) {
        dashboards(
            where: {
            slug: {_eq: $slug}, 
            _or: [
                    {user: {name: {_eq: $user}}}, 
                    {team: {handle: {_eq: $user}}}
                ]
            }
        ) {
        ...Dashboard
        favorite_dashboards(where: {user_id: $session_id}, limit: 1) {
          created_at
        }
      }
    }
    
    fragment Dashboard on dashboards {
      id
      name
      slug
      is_private
      is_archived
      created_at
      updated_at
      tags
      user {
      ...User 
      }
      team {
      ...Team
      }
    forked_dashboard {    
        slug    
        name    
        user {      
            name      
        }  
        team {
          handle
        }
    }
      text_widgets {
        id
        created_at
        updated_at
        text
        options
      }
      visualization_widgets {
        id
        created_at
        updated_at
        options
        visualization {
          ...Visualization
        }
      }
      param_widgets {
        id
        key
        visualization_widget_id
        query_id
        dashboard_id
        options
        created_at
        updated_at
      }
      dashboard_favorite_count_all {
        favorite_count
      }
      trending_scores {
        score_1h
        score_4h
        score_24h
        updated_at  
      }
      }
    fragment User on users {
      id
      name
      profile_image_url
      }
    fragment Team on teams {
      id
      name
      handle
      profile_image_url
      }
    fragment Visualization on visualizations {
      id
      type
      name
      options
      created_at
      query_details {
        query_id
        name
        description
        show_watermark
        parameters
        user {
            id
            name
            profile_image_url      
        }
        team {
          id
          name
          handle
          profile_image_url
        }  
      }
    }
"""

FIND_QUERY_POST = """
    query FindQuery(
        $session_id: Int_comparison_exp, 
        $id: Int!, 
        $favs_last_24h: Boolean! = false, 
        $favs_last_7d: Boolean! = false, 
        $favs_last_30d: Boolean! = false, 
        $favs_all_time: Boolean! = true
    ) {
      queries(where: {id: {_eq: $id}}) {
        ...Query
        favorite_queries(where: {user_id: $session_id}, limit: 1) {
          created_at
        }
      }
    }
    
    fragment Query on queries {
      ...BaseQuery
      ...QueryVisualizations
      ...QueryForked
      ...QueryUsers
      ...QueryTeams
      ...QueryFavorites
    }
    
    fragment BaseQuery on queries {
      id
      dataset_id
      name
      description
      query
      is_private
      is_temp
      is_archived
      created_at
      updated_at
      schedule
      tags
      parameters
    }
    
    fragment QueryVisualizations on queries {
      visualizations {
        id
        type
        name
        options
        created_at
      }
    }
    
    fragment QueryForked on queries {
      forked_query {
        id
        name
        user {
          name
        }
        team {
          handle
        }
      }
    }
    
    fragment QueryUsers on queries {
      user {
        ...User
      }
      team {
        id
        name
        handle
        profile_image_url
      }
    }
    
    fragment User on users {
      id
      name
      profile_image_url
    }
    fragment QueryTeams on queries {
      team {
        ...Team
      }
    }
    fragment Team on teams {
      id
      name
      handle
      profile_image_url
    }
    fragment QueryFavorites on queries {
      query_favorite_count_all @include(if: $favs_all_time) {
        favorite_count
      }
      query_favorite_count_last_24h @include(if: $favs_last_24h) {
        favorite_count
      }
      query_favorite_count_last_7d @include(if: $favs_last_7d) {
        favorite_count
      }
      query_favorite_count_last_30d @include(if: $favs_last_30d) {
        favorite_count
      }
    }
"""



"""Utility code for I/O related tasks"""
import csv
import os
from dataclasses import fields, astuple, dataclass, is_dataclass
from typing import Any

FILE_OUT_PATH = os.environ.get("FILE_OUT_PATH", "./out")


@dataclass
class File:
    """Simple structure for declaring and passing around filenames"""

    name: str
    path: str = FILE_OUT_PATH

    def filename(self) -> str:
        """Returns the complete path to file"""
        return os.path.join(self.path, self.name)

    def __str__(self) -> str:
        return self.filename()


def write_to_csv(data_list: list[Any], outfile: File) -> None:
    """Writes `data_list` to `filename` as csv"""
    print(f"dumping {len(data_list)} results to {outfile.name}")

    # Creates out path if it doesn't exist.
    if not os.path.exists(outfile.path):
        os.makedirs(outfile.path)

    with open(outfile.filename(), "w", encoding="utf-8") as out_file:
        if len(data_list) == 0:
            return
        sample = data_list[0]
        assert is_dataclass(sample), "Method only accepts lists of type dataclass"
        headers = [f.name for f in fields(sample)]
        data_tuple = [astuple(x) for x in data_list]

        dict_writer = csv.DictWriter(out_file, headers, lineterminator="\n")
        dict_writer.writeheader()
        writer = csv.writer(out_file, lineterminator="\n")
        writer.writerows(data_tuple)



import datetime
import json
import unittest

from duneapi.types import Network, MetaData, QueryResults, QueryParameter, Address


class TestAddress(unittest.TestCase):
    def setUp(self) -> None:
        self.lower_case_address = "0xde1c59bc25d806ad9ddcbe246c4b5e5505645718"
        self.check_sum_address = "0xDEf1CA1fb7FBcDC777520aa7f396b4E015F497aB"
        self.invalid_address = "0x12"
        self.dune_format = "\\x5d4020b9261f01b6f8a45db929704b0ad6f5e9e6"

    def test_invalid(self):
        with self.assertRaises(ValueError) as err:
            Address(address=self.invalid_address)
        self.assertEqual(
            str(err.exception), f"Invalid Ethereum Address {self.invalid_address}"
        )

    def test_valid(self):
        self.assertEqual(
            Address(address=self.lower_case_address).address,
            "0xdE1c59Bc25D806aD9DdCbe246c4B5e5505645718",
        )
        self.assertEqual(
            Address(address=self.check_sum_address).address,
            "0xDEf1CA1fb7FBcDC777520aa7f396b4E015F497aB",
        )
        self.assertEqual(
            Address(address=self.dune_format).address,
            "0x5d4020b9261F01B6f8a45db929704b0Ad6F5e9E6",
        )

    def test_inequality(self):
        zero = Address("0x0000000000000000000000000000000000000000")
        one = Address("0x1000000000000000000000000000000000000000")
        a_lower = Address("0xa000000000000000000000000000000000000000")
        a_upper = Address("0xA000000000000000000000000000000000000000")
        b_lower = Address("0xb000000000000000000000000000000000000000")
        c_upper = Address("0xC000000000000000000000000000000000000000")
        self.assertLess(zero, one)
        self.assertLess(one, a_lower)
        self.assertEqual(a_lower, a_upper)
        self.assertLess(a_upper, b_lower)
        self.assertLess(b_lower, c_upper)


class TestNetworkEnum(unittest.TestCase):
    def test_string_rep(self):
        self.assertEqual(str(Network.MAINNET), "Ethereum Mainnet")
        self.assertEqual(str(Network.GCHAIN), "Gnosis Chain")
        self.assertEqual(str(Network.SOLANA), "Solana")
        self.assertEqual(str(Network.BINANCE), "Binance Smart Chain")
        self.assertEqual(str(Network.POLYGON), "Polygon")

    def test_try_from_string(self):
        # Mainnet
        self.assertEqual(Network.MAINNET, Network.from_string("mainnet"))
        self.assertEqual(Network.MAINNET, Network.from_string("MaInNet"))
        self.assertEqual(Network.MAINNET, Network.from_string("ethereum mainnet"))
        self.assertEqual(Network.MAINNET, Network.from_string("ETHERIUM MaInNet"))
        self.assertEqual(Network.MAINNET, Network.from_string("MaInNet ethereum"))

        # Gnosis Chain
        self.assertEqual(Network.GCHAIN, Network.from_string("gchain"))
        self.assertEqual(Network.GCHAIN, Network.from_string("Gnosis Chain"))
        self.assertEqual(Network.GCHAIN, Network.from_string("GNOChain"))

        # Binance
        self.assertEqual(Network.BINANCE, Network.from_string("bsc"))
        self.assertEqual(Network.BINANCE, Network.from_string("BSC"))
        self.assertEqual(Network.BINANCE, Network.from_string("Binance"))

        # Optimism
        self.assertEqual(Network.OPTIMISM_V1, Network.from_string("optimism v1"))
        self.assertEqual(Network.OPTIMISM_V2, Network.from_string("optimism v2"))

        # No match!
        with self.assertRaises(ValueError) as err:
            Network.from_string("toot")
        self.assertEqual("could not parse Network from 'toot'", str(err.exception))


class TestQueryResults(unittest.TestCase):
    def setUp(self) -> None:
        self.metadata_content = {
            "id": "3158cc2c-5ed1-4779-b523-eeb9c3b34b21",
            "job_id": "093e440d-66ce-4c00-81ec-2406f0403bc0",
            "error": None,
            "runtime": 0,
            "generated_at": "2022-03-19T07:11:37.344998+00:00",
            "columns": ["number", "size", "time", "block_hash", "tx_fees"],
            "__typename": "query_results",
        }
        self.valid_empty_results = {
            "query_results": [self.metadata_content],
            "get_result_by_job_id": [],
            "query_errors": [],
        }

    def test_metadata_constructor(self):
        result = MetaData(json.dumps(self.metadata_content))
        self.assertEqual(result.__dict__, self.metadata_content)

    def test_constructor_success(self):
        results = QueryResults(self.valid_empty_results)
        self.assertEqual(results.data, [])

    def test_constructor_assertions(self):
        with self.assertRaises(AssertionError) as err:
            QueryResults({"a": [{}]})
        self.assertEqual(str(err.exception), "invalid keys dict_keys(['a'])")

        invalid_query_results = {
            "query_results": [self.metadata_content, {}],  # Not of list type!
            "get_result_by_job_id": [],
            "query_errors": [],
        }
        with self.assertRaises(AssertionError) as err:
            QueryResults(invalid_query_results)
        self.assertEqual(
            str(err.exception),
            f"Unexpected query_results {invalid_query_results}",
        )


class TestQueryParameter(unittest.TestCase):
    def test_constructors_and_to_dict(self):
        number_type = QueryParameter.number_type("Number", 1)
        text_type = QueryParameter.text_type("Text", "hello")
        date_type = QueryParameter.date_type("Date", datetime.datetime(2022, 3, 10))

        self.assertEqual(
            number_type.to_dict(), {"key": "Number", "type": "number", "value": "1"}
        )
        self.assertEqual(
            text_type.to_dict(), {"key": "Text", "type": "text", "value": "hello"}
        )
        self.assertEqual(
            date_type.to_dict(),
            {"key": "Date", "type": "datetime", "value": "2022-03-10 00:00:00"},
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock, Mock

from duneapi.api import DuneAPI
from duneapi.types import DuneQuery, Network


class TestDuneAnalytics(unittest.TestCase):
    def setUp(self) -> None:
        self.dune = DuneAPI("user", "password")
        self.query = DuneQuery(
            raw_sql="",
            description="",
            network=Network.MAINNET,
            query_id=0,
            parameters=[],
            name="Test",
        )

    def test_retry(self):
        self.dune.execute_and_await_results = MagicMock(return_value=1)
        self.dune.initiate_query = MagicMock(return_value=None)
        self.dune.open_query = MagicMock(return_value="")
        self.dune.max_retries = 0
        with self.assertRaises(Exception):
            self.dune.fetch(self.query)

        self.dune.max_retries = 1
        self.assertEqual(self.dune.fetch(self.query), 1)

        self.dune.execute_and_await_results = Mock(side_effect=Exception("Max retries"))
        with self.assertRaises(Exception):
            self.dune.fetch(self.query)


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime

from duneapi.util import datetime_parser, open_query, duplicates, DUNE_DATE_FORMAT


class TestUtilities(unittest.TestCase):
    def test_date_parser(self):
        # def datetime_parser(dct: dict[str, Any]) -> dict[str, Any]:
        no_dates = {"a": 1, "b": "hello", "c": [None, 3.14]}
        self.assertEqual(datetime_parser(no_dates), no_dates)

        valid_date_str = "1985-03-10 05:00:00"
        invalid_date_str = "1985/03/10 - literally anything else"
        with_date = {"valid_date": valid_date_str, "invalid_date": invalid_date_str}
        self.assertEqual(
            datetime_parser(with_date),
            {
                "valid_date": datetime.strptime(valid_date_str, DUNE_DATE_FORMAT),
                "invalid_date": invalid_date_str,
            },
        )

    def test_open_query(self):
        query = "select 10 - '{{IntParameter}}' as value"
        self.assertEqual(query, open_query("./tests/queries/test_query.sql"))

    def test_duplicates(self):
        self.assertEqual(duplicates([1, 2]), [])
        self.assertEqual(duplicates([1, 2, 2, 4]), [2])
        self.assertEqual(duplicates([1, 1, 2, 2]), [1, 2])

        with self.assertRaises(TypeError) as err:
            duplicates([{"x": 1, "y": 2}])


if __name__ == "__main__":
    unittest.main()



from duneapi.api import DuneAPI
from duneapi.dashboard import DuneDashboard


if __name__ == "__main__":
    DuneDashboard.from_dune(
        api=DuneAPI.new_from_environment(), dashboard_slug="Demo-Dashboard"
    )



import unittest

from duneapi.api import DuneAPI
from duneapi.types import Network, QueryParameter, DuneQuery


class TestDuneAnalytics(unittest.TestCase):
    def setUp(self) -> None:
        self.five = 5
        self.one = 1
        self.parameter_name = "IntParameter"
        self.column_name = "value"
        self.dune = DuneAPI.new_from_environment()
        self.mainnet_query = self.network_query(Network.MAINNET)

    def network_query(self, network: Network) -> DuneQuery:
        return DuneQuery.from_environment(
            # Note that consecutive double brace brackets in formatted strings
            # become single brace brackets, so this query is
            # select 5 - '{{IntParameter}}' as value
            raw_sql=f"select {self.five} - '{{{{{self.parameter_name}}}}}' as {self.column_name}",
            description="Test Description",
            network=network,
            parameters=[QueryParameter.number_type(self.parameter_name, self.one)],
            name="Test Fetch",
        )

    def test_initiate_query(self):
        self.assertEqual(True, self.dune.initiate_query(self.mainnet_query))

    def test_execute_query(self):
        q_id = self.mainnet_query.query_id
        params = self.mainnet_query.parameters
        self.assertNotEqual(None, self.dune.execute(q_id, params))

    def test_interface(self):
        """
        This test indirectly touches all of
        - fetch
        - initiate_new_query
        - execute_and_await_results
        - execute_query
        - post_dune_request
        - Tests that the API works on all supported "Networks"
        essentially all the methods of the API
        """
        dune = DuneAPI.new_from_environment()
        for network in Network:
            query = self.network_query(network)
            res = dune.fetch(query)
            self.assertEqual(len(res), 1)
            self.assertEqual(res[0][self.column_name], self.five - self.one)


if __name__ == "__main__":
    unittest.main()



from duneapi.api import DuneAPI
from duneapi.dashboard import DuneDashboard

dune = DuneAPI.new_from_environment()
dashboard = DuneDashboard.from_file(
    api=dune, filename="./example/dashboard/_config.json"
)
dashboard.update()
print("Updated", dashboard)



import unittest

from duneapi.api import DuneAPI
from duneapi.dashboard import DuneDashboard


class TestDashboard(unittest.TestCase):
    def test_load(self):
        configured_dashboard = DuneDashboard.from_file(
            api=DuneAPI.new_from_environment(),
            filename="./example/dashboard/_config.json",
        )
        configured_dashboard.update()
        fetched_dashboard = DuneDashboard.from_dune(
            api=DuneAPI.new_from_environment(),
            dashboard_slug="Demo-Dashboard",
            save_config=False,
        )
        self.assertEqual(fetched_dashboard, configured_dashboard)


if __name__ == "__main__":
    unittest.main()



"""
A collection of types associated with Dune Analytics API.

Includes also, the base classes for a Dune queries, parameters and Request Post Data
All operations/routes available for interaction with Dune API - looks like graphQL
"""
from __future__ import annotations

import json
import os
import re
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Collection, Optional
from web3 import Web3

from dotenv import load_dotenv

from .logger import set_log
from .util import datetime_parser, open_query, postgres_date

log = set_log(__name__)

PostData = dict[str, Collection[str]]
# key_map = {"outer1": {"inner11", "inner12}, "outer2": {"inner21"}}
KeyMap = dict[str, set[str]]

ListInnerResponse = dict[str, list[dict[str, dict[str, str]]]]
DictInnerResponse = dict[str, dict[str, Any]]

DuneRecord = dict[str, str]

# pylint: disable=too-few-public-methods
class Address:
    """
    Class representing Ethereum Address as a hexadecimal string of length 42.
    The string must begin with '0x' and the other 40 characters
    are digits 0-9 or letters a-f. Upon creation (from string) addresses
    are validated and stored in their check-summed format.
    """

    def __init__(self, address: str):
        # Dune uses \x instead of 0x (i.e. bytea instead of hex string)
        # This is just a courtesy to query writers,
        # so they don't have to convert all addresses to hex strings manually
        address = address.replace("\\x", "0x")
        if Address._is_valid(address):
            self.address: str = Web3.toChecksumAddress(address)
        else:
            raise ValueError(f"Invalid Ethereum Address {address}")

    def __str__(self) -> str:
        return str(self.address)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Address):
            return self.address == other.address
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Address):
            return str(self).lower() < str(other).lower()
        return False

    def __hash__(self) -> int:
        return self.address.__hash__()

    @classmethod
    def zero(cls) -> Address:
        """Returns Null Ethereum Address"""
        return cls("0x0000000000000000000000000000000000000000")

    @classmethod
    def from_int(cls, num: int) -> Address:
        """
        Construct an address from int.
        Used for testing, so that 123 -> "0x0000000000000000000000000000000000000123"
        """
        return cls(f"0x{str(num).rjust(40, '0')}")

    @staticmethod
    def _is_valid(address: str) -> bool:
        match_result = re.match(
            pattern=r"^(0x)?[0-9a-f]{40}$", string=address, flags=re.IGNORECASE
        )
        return match_result is not None


# pylint: disable=too-few-public-methods
# TODO - use namedtuple for MetaData and QueryResults
class MetaData:
    """The standard information returned from the Dune API as `query_results`"""

    id: str
    job_id: str
    error: Optional[str]
    runtime: int
    generated_at: datetime
    columns: list[str]

    def __init__(self, obj: str):
        """
        Constructor method
        :param obj: input should have the following form

        Example input:
        {
            'id': '3158cc2c-5ed1-4779-b523-eeb9c3b34b21',
            'job_id': '093e440d-66ce-4c00-81ec-2406f0403bc0',
            'error': None,
            'runtime': 0,
            'generated_at': '2022-03-19T07:11:37.344998+00:00',
            'columns': ['number', 'size', 'time', 'block_hash', 'tx_fees'],
            '__typename': 'query_results'
        }
        """
        self.__dict__ = json.loads(obj, object_hook=datetime_parser)


class QueryResults:
    """Class containing the Data results of a Dune Select Query"""

    meta: Optional[MetaData]
    data: list[DuneRecord]

    def __init__(self, data: ListInnerResponse):
        assert data.keys() == {
            "query_results",
            "get_result_by_job_id",
            "query_errors",
        }, f"invalid keys {data.keys()}"
        assert len(data["query_results"]) == 1, f"Unexpected query_results {data}"
        # Could wrap meta conversion into a try-catch, since we don't really need it.
        # But, I can't think of a broad enough exception that won't trip up the liner.
        self.meta = MetaData(json.dumps(data["query_results"][0]))

        self.data = [rec["data"] for rec in data["get_result_by_job_id"]]


class Network(Enum):
    """Enum for supported EVM networks"""

    SOLANA = 1
    MAINNET = 4
    GCHAIN = 6
    POLYGON = 7
    OPTIMISM_V1 = 8
    BINANCE = 9
    OPTIMISM_V2 = 10

    def __str__(self) -> str:
        string_map = {
            Network.SOLANA: "Solana",
            Network.MAINNET: "Ethereum Mainnet",
            Network.GCHAIN: "Gnosis Chain",
            Network.POLYGON: "Polygon",
            Network.OPTIMISM_V1: "Optimism (OVM 1.0)",
            Network.OPTIMISM_V2: "Optimism (OVM 2.0)",
            Network.BINANCE: "Binance Smart Chain",
        }
        return string_map.get(self, "Unknown Network")

    @classmethod
    def from_string(cls, network_str: str) -> Network:
        """
        Attempts to parse network name from string.
        returns None is no match
        """
        patterns = {
            r"(.*)mainnet": cls.MAINNET,
            r"g(.*)chain": cls.GCHAIN,
            r"solana": cls.SOLANA,
            r"poly": cls.POLYGON,
            r"optimism(.*)1": cls.OPTIMISM_V1,
            r"optimism(.*)2": cls.OPTIMISM_V2,
            r"bsc": cls.BINANCE,
            r"binance": cls.BINANCE,
        }
        for pattern, network in patterns.items():
            if re.match(pattern, network_str, re.IGNORECASE):
                return network
        raise ValueError(f"could not parse Network from '{network_str}'")


class ParameterType(Enum):
    """
    Enum of the 4 distinct dune parameter types
    """

    TEXT = "text"
    NUMBER = "number"
    DATE = "datetime"
    ENUM = "enum"

    @classmethod
    def from_string(cls, type_str: str) -> ParameterType:
        """
        Attempts to parse Parameter from string.
        returns None is no match
        """
        patterns = {
            r"text": cls.TEXT,
            r"number": cls.NUMBER,
            r"date": cls.DATE,
            r"enum": cls.ENUM,
        }
        for pattern, param in patterns.items():
            if re.match(pattern, type_str, re.IGNORECASE):
                return param
        raise ValueError(f"could not parse Network from '{type_str}'")


class QueryParameter:
    """Class whose instances are Dune Compatible Query Parameters"""

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        value: Any,
        options: Optional[list[str]] = None,
    ):
        if not options:
            options = []
        self.key: str = name
        self.type: ParameterType = parameter_type
        self.value = value
        self.options = options

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QueryParameter):
            return NotImplemented
        return all(
            [
                self.key == other.key,
                self.value == other.value,
                self.type.value == other.type.value,
            ]
        )

    @classmethod
    def text_type(cls, name: str, value: str) -> QueryParameter:
        """Constructs a Query parameter of type text"""
        return cls(name, ParameterType.TEXT, value)

    @classmethod
    def number_type(cls, name: str, value: int | float) -> QueryParameter:
        """Constructs a Query parameter of type number"""
        return cls(name, ParameterType.NUMBER, value)

    @classmethod
    def date_type(cls, name: str, value: datetime | str) -> QueryParameter:
        """
        Constructs a Query parameter of type date.
        For convenience, we allow proper datetime type, or string
        """
        if isinstance(value, str):
            value = postgres_date(value)
        return cls(name, ParameterType.DATE, value)

    @classmethod
    def enum_type(cls, name: str, value: str, options: list[str]) -> QueryParameter:
        """Constructs a Query parameter of type number"""
        return cls(name, ParameterType.ENUM, value, options)

    def _value_str(self) -> str:
        if self.type in (ParameterType.TEXT, ParameterType.NUMBER, ParameterType.ENUM):
            return str(self.value)
        if self.type == ParameterType.DATE:
            # This is the postgres string format of timestamptz
            return str(self.value.strftime("%Y-%m-%d %H:%M:%S"))
        raise TypeError(f"Type {self.type} not recognized!")

    def to_dict(self) -> dict[str, str | list[str]]:
        """Converts QueryParameter into string json format accepted by Dune API"""
        results: dict[str, str | list[str]] = {
            "key": self.key,
            "type": self.type.value,
            "value": self._value_str(),
        }
        if self.type == ParameterType.ENUM:
            results["enumOptions"] = self.options
        return results

    @classmethod
    def from_dict(cls, obj: dict[str, Any]) -> QueryParameter:
        """
        Constructs Query Parameters from json.
        TODO - this could probably be done similar to the __init__ method of MetaData
        """
        name, value = obj["key"], obj["value"]
        p_type = ParameterType.from_string(obj["type"])
        if p_type == ParameterType.DATE:
            return cls.date_type(name, value)
        if p_type == ParameterType.TEXT:
            assert isinstance(value, str)
            return cls.text_type(name, value)
        if p_type == ParameterType.NUMBER:
            if isinstance(value, str):
                value = float(value) if "." in value else int(value)
            return cls.number_type(name, value)
        if p_type == ParameterType.ENUM:
            return cls.enum_type(name, value, obj["enumOptions"])
        raise ValueError(f"Could not parse Query parameter from {obj}")

    def __str__(self) -> str:
        return (
            f"QueryParameter("
            f"name: {self.key}, "
            f"value: {self.value}, "
            f"type: {self.type.value})"
        )


@dataclass
class Post:
    """Holds query json and response validation details"""

    data: PostData
    key_map: KeyMap


@dataclass
class DashboardTile:
    """
    A slightly different arrangement of data that is essentially equivalent to a Query
    Acts as an intermediary type when composing queries from json
    """

    name: str
    description: str
    select_file: str
    query_id: int
    network: Network
    parameters: list[QueryParameter]
    base_file: Optional[str]

    @classmethod
    def from_dict(cls, obj: dict[str, Any], path: str) -> DashboardTile:
        """Constructs Record from Dune Data as string dict"""
        return cls(
            name=obj.get("name", "untitled"),
            description=obj.get("description", ""),
            select_file="/".join([path, obj["query_file"]]),
            network=Network.from_string(obj["network"]),
            query_id=int(obj["id"]),
            parameters=[QueryParameter.from_dict(p) for p in obj.get("parameters", [])],
            base_file=obj.get("requires"),
        )

    def build_query(self) -> str:
        """Constructs a query from base file and select file attributes"""
        if self.base_file is not None:
            components = map(open_query, [self.base_file, self.select_file])
            return "\n".join(list(components))
        return open_query(self.select_file)


@dataclass
class DuneQuery:
    """Contains all the relevant data necessary to initiate a Dune Query"""

    query_id: int
    raw_sql: str = ""
    name: str = ""
    description: str = ""
    network: Network = Network.MAINNET
    parameters: list[QueryParameter] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(self.query_id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DuneQuery):
            return NotImplemented
        equality_conditions = [
            self.name == other.name,
            self.description == other.description,
            self.raw_sql == other.raw_sql,
            self.network.value == other.network.value,
            self.query_id == other.query_id,
            self.parameters == other.parameters,
        ]
        log.debug(f"Equality Conditions: {equality_conditions}")
        return all(equality_conditions)

    @classmethod
    def from_environment(
        cls,
        raw_sql: str,
        network: Network,
        description: str = "",
        parameters: Optional[list[QueryParameter]] = None,
        name: Optional[str] = None,
    ) -> DuneQuery:
        """Constructs a query using the Universal Query ID provided in env file."""
        load_dotenv()
        return cls(
            raw_sql=raw_sql,
            description=description,
            network=network,
            parameters=parameters if parameters is not None else [],
            name=name if name else "untitled",
            query_id=int(os.environ["DUNE_QUERY_ID"]),
        )

    @classmethod
    def from_tile(cls, tile: DashboardTile) -> DuneQuery:
        """Constructs Dune Query from DashboardTile object"""
        return cls(
            name=tile.name,
            description=tile.description,
            raw_sql=tile.build_query(),
            network=tile.network,
            parameters=tile.parameters,
            query_id=tile.query_id,
        )

    def _request_parameters(self) -> list[dict[str, str | list[str]]]:
        return [p.to_dict() for p in self.parameters]

    def upsert_query_post(self) -> Post:
        """Returns json data for a post of type UpsertQuery"""
        object_data: dict[str, Any] = {
            "id": self.query_id,
            "schedule": None,
            "dataset_id": self.network.value,
            "name": self.name,
            "query": self.raw_sql,
            "user_id": 84,
            "description": self.description,
            "is_archived": False,
            "is_temp": False,
            "tags": [],
            "parameters": self._request_parameters(),
            "visualizations": {
                "data": [],
                "on_conflict": {
                    "constraint": "visualizations_pkey",
                    "update_columns": ["name", "options"],
                },
            },
        }
        key_map = {
            "insert_queries_one": {
                "id",
                "dataset_id",
                "name",
                "description",
                "query",
                "is_private",
                "is_temp",
                "is_archived",
                "created_at",
                "updated_at",
                "schedule",
                "tags",
                "parameters",
                "visualizations",
                "forked_query",
                "user",
                "team",
                "query_favorite_count_all",
                "favorite_queries",
            }
        }
        return Post(
            data={
                "operationName": "UpsertQuery",
                "variables": {
                    "object": object_data,
                    "on_conflict": {
                        "constraint": "queries_pkey",
                        "update_columns": [
                            "dataset_id",
                            "name",
                            "description",
                            "query",
                            "schedule",
                            "is_archived",
                            "is_temp",
                            "tags",
                            "parameters",
                        ],
                    },
                    "session_id": 0,  # must be an int, but value is irrelevant
                },
                "query": """
                mutation UpsertQuery(
                  $session_id: Int!
                  $object: queries_insert_input!
                  $on_conflict: queries_on_conflict!
                  $favs_last_24h: Boolean! = false
                  $favs_last_7d: Boolean! = false
                  $favs_last_30d: Boolean! = false
                  $favs_all_time: Boolean! = true
                ) {
                  insert_queries_one(object: $object, on_conflict: $on_conflict) {
                    ...Query
                    favorite_queries(where: { user_id: { _eq: $session_id } }, limit: 1) {
                      created_at
                    }
                  }
                }
                fragment Query on queries {
                  ...BaseQuery
                  ...QueryVisualizations
                  ...QueryForked
                  ...QueryUsers
                  ...QueryTeams
                  ...QueryFavorites
                }
                fragment BaseQuery on queries {
                  id
                  dataset_id
                  name
                  description
                  query
                  is_private
                  is_temp
                  is_archived
                  created_at
                  updated_at
                  schedule
                  tags
                  parameters
                }
                fragment QueryVisualizations on queries {
                  visualizations {
                    id
                    type
                    name
                    options
                    created_at
                  }
                }
                fragment QueryForked on queries {
                  forked_query {
                    id
                    name
                    user {
                      name
                    }
                    team {
                     handle
                    }
                  }
                }
                fragment QueryUsers on queries {
                  user {
                    ...User
                  }
                  team {
                    id
                    name
                    handle
                    profile_image_url
                  }
                }
                fragment User on users {
                  id
                  name
                  profile_image_url
                }
                fragment QueryTeams on queries {
                  team {
                    ...Team
                  }
                }
                fragment Team on teams {
                  id
                  name
                  handle
                  profile_image_url
                }
                fragment QueryFavorites on queries {
                  query_favorite_count_all @include(if: $favs_all_time) {
                    favorite_count
                  }
                  query_favorite_count_last_24h @include(if: $favs_last_24h) {
                    favorite_count
                  }
                  query_favorite_count_last_7d @include(if: $favs_last_7d) {
                    favorite_count
                  }
                  query_favorite_count_last_30d @include(if: $favs_last_30d) {
                    favorite_count
                  }
                }
                """,
            },
            key_map=key_map,
        )

    @staticmethod
    def find_result_by_job(job_id: str) -> Post:
        """Returns json data for a post of type FindResultDataByResult"""
        query = """
        query FindResultDataByJob($job_id: uuid!) {
          query_results(where: {job_id: {_eq: $job_id}}) {
            id
            job_id
            runtime
            generated_at
            columns
          }
          query_errors(where: {job_id: {_eq: $job_id}}) {
            id
            job_id
            runtime
            message
            metadata
            type
            generated_at
          }
          get_result_by_job_id(args: {want_job_id: $job_id}) {
            data
          }
        }
        """
        return Post(
            data={
                "operationName": "FindResultDataByJob",
                "variables": {"job_id": job_id},
                "query": query,
            },
            key_map={
                "query_results": {
                    "id",
                    "job_id",
                    "runtime",
                    "generated_at",
                    "columns",
                },
                "query_errors": {
                    "id",
                    "job_id",
                    "runtime",
                    "message",
                    "metadata",
                    "type",
                    "generated_at",
                },
                "get_result_by_job_id": {"data"},
            },
        )

    @staticmethod
    def get_queue_position(job_id: str) -> Post:
        """Returns json data for a post of type GetQueuePosition
        This is meant to determine when query execution has completed.
        """
        query = """
        query GetQueuePosition($job_id: uuid!) {
          get_queue_position(job_id: $job_id) {
            pos
          }
          jobs_by_pk(id: $job_id) {
            id
            user_id
            category
            created_at
            locked_until
          }
        }
        """
        return Post(
            data={
                "operationName": "GetQueuePosition",
                "variables": {"job_id": job_id},
                "query": query,
            },
            key_map={"data": {"view_queue_positions", "jobs_by_pk"}},
        )


def execute_query_post_data(query_id: int, params: list[QueryParameter]) -> Post:
    """Returns json data for a post of type ExecuteQuery"""
    return Post(
        data={
            "operationName": "ExecuteQuery",
            "variables": {
                "query_id": query_id,
                "parameters": [p.to_dict() for p in params],
            },
            "query": """
                mutation ExecuteQuery($query_id: Int!, $parameters: [Parameter!]!) {
                  execute_query(query_id: $query_id, parameters: $parameters) {
                    job_id
                  }
                }""",
        },
        key_map={"execute_query": {"job_id"}},
    )
